from scipy.interpolate import interp1d


def gap_speed_factor(grade):
  """Calculate Strava's GAP speed-factor as a function of percent grade.

  The factor will be greater than 1.0 if GAP is faster than horizontal speed,
  and less than 1.0 if GAP is slower than horizontal speed.

  This function is the result of an experiment I performed with Strava.
  I spoofed activity files to upload to Strava to reverse-engineer
  their formula for Grade Adjusted Pace (GAP). The files each depicted a
  runner moving at a constant speed up a constant incline. I then took a ratio
  of the grade-adjusted speeds to horizontal speeds at a variety of speed and 
  grade combinations.

  I found that Strava applies a pace adjustment factor that is dependent only
  on grade (ie independent of speed).

  Grade is constrained to (-45%, 45%), because that was the range of grades
  in my investigation. That range also coincides with the values investigated
  by Minetti (2002). Rather than extrapolating outside of that range, I chose
  to cap the adjustment factor.
  
  Strava's GAP is now a HR-based, big data-derived quantity, so I don't know
  how reasonable it is at the steepest grades; there are certainly fewer data
  points to go on at the extreme ends of the range. In other analyses I've
  done, I don't think I have seen Strava report a grade steeper than 50%, 
  so it is possible that defines its range of the GAP formula's validity. 
  Strava seems to smooth its elevation and/or grade time series to fit into
  this range.

  Args:
    grade (float): Decimal grade to evaluate the GAP speed-factor.
  Returns:
    float: Factor that converts speed to grade-adjusted speed.
  """

  adjustment_factors = {
    45: 4.286,
    30: 3.158,
    25: 2.727,
    20: 2.297,
    15: 1.846,
    10: 1.459,
    8: 1.337,
    6: 1.228,
    4: 1.135,
    2: 1.055,
    0: 1.0,
    -2: 0.96,
    -4: 0.918,
    -6: 0.891,
    -8: 0.876,
    -10: 0.876,
    -15: 0.941,
    -20: 1.081,
    -25: 1.273,
    -30: 1.495,
    -45: 2.096
  }

  interp_fn = interp1d(
    list(adjustment_factors.keys()),
    list(adjustment_factors.values())
  )

  return interp_fn(grade * 100)



def poly_5(x, a, b, c, d, e, f):
  """Generic 5th-order polynomial function.
  
  Args:
    x (float or array(float)): the quantity to apply the polynomial to.
      Could be a single value, a numpy array, or a pandas Series.
    a (float): coefficient for the fifth-order term.
    b (float): coefficient for the fourth-order term.
    c (float): coefficient for the third-order term.
    d (float): coefficient for the second-order term.
    e (float): coefficient for the first-order term.
    f (float): coefficient for the zeroth-order term.

  Returns:
    float
  """
  return a * x ** 5 + b * x ** 4 + c * x ** 3 + d * x ** 2 + e * x + f


def cost_of_running(decimal_grade):
  """Energy consumption of running (per kg per m), according Minetti 2002.

  This is the curve fit supplied by the authors of the paper.

  Args:
    decimal_grade (float): decimal grade of terrain, i.e. 0.2 for 20%.
    equation (str): 
  Returns:
    float: the estimated cost of running, in J/kg/m, at the given grade.

  TODO:
    * Constrain the algorithm to the (+/-)45% range
  """
  return poly_5(decimal_grade, 155.4, -30.4, -43.3, 46.3, 19.5, 3.6)


def cost_of_walking(decimal_grade):
  """Energy consumption of walking (per kg per m), according Minetti 2002.

  This is the curve fit supplied by the authors of the paper.

  NOTE: Energy consumption varied with walking speed, and this equation
  represents the minimum energy consumption as a function of grade.

  Args:
    decimal_grade (float): decimal grade of terrain, i.e. 0.2 for 20%.
    equation (str): 
  Returns:
    float: the estimated cost of walking, in J/kg/m, at the given grade.
  """
  return poly_5(decimal_grade, 280.5, -58.7, -76.8, 51.9, 19.6, 2.5)


"""Formulas that come directly from another source.

Examples:
  - the O2 consumption curve from Daniels Running Formula
  - the adjustment factor that Strava applies to speeds based on grade
    to come up with GAP.
"""



from scipy.interpolate import interp1d


def ngp_speed_factor(grade):
  """Calculate TrainingPeaks' NGP pace-factor as a function of percent grade.

  The factor will be greater than 1.0 if NGP is faster than horizontal speed,
  and less than 1.0 if NGP is slower than horizontal speed.

  Grade is constrained to (-25%, 30%), because that was the range of grades
  that yielded reasonable paces in my investigation:
    * 0.45 adjusted 8:00 pace to 0:17 pace.
    * The adjustment factor jumped wildly from -26% to -27% and from
      31% to 32%.

  Rather than extrapolating outside of that range, I chose to cap the 
  adjustment factor.

  TrainingPeaks does not provide any info about how they derived their
  grade adjustment factor, but the shape of the curve sure looks a lot
  like Strava's GAP.

  Args:
    grade (float): Decimal grade to evaluate the NGP speed-factor.
  Returns:
    float: Factor that converts speed to NGP adjusted speed.
  """

  adjustment_factors = {
    45: 28.235,
    32: 20.0,
    31: 3.556,
    30: 3.429,
    25: 2.874,
    20: 2.365,
    15: 1.905,
    10: 1.514,
    8: 1.383,
    6: 1.266,
    4: 1.162,
    2: 1.074,
    0: 1.0,
    -2: 0.941,
    -4: 0.897,
    -6: 0.87,
    -8: 0.856,
    -10: 0.857,
    -15: 0.921,
    -20: 1.067,
    -25: 1.28,
    -26: 1.33,
    -27: 2.051,
    -30: 2.297,
    -45: 3.934
  }

  interp_fn = interp1d(
    list(adjustment_factors.keys()),
    list(adjustment_factors.values())
  )

  return interp_fn(grade * 100)


import os
from setuptools import setup, find_packages


def read(rel_path):
  """Read a file so python does not have to import it.
  
  Inspired by (taken from) pip's `setup.py`.
  """
  here = os.path.abspath(os.path.dirname(__file__))
  with open(os.path.join(here, rel_path), 'r') as fp:
    return fp.read()


def get_version(rel_path):
  """Manually read through a file to retrieve its `__version__`.
  
  Inspired by (taken from) pip's `setup.py`.
  """
  for line in read(rel_path).splitlines():
    if line.startswith('__version__'):
      delim = "'" if "'" in line else '"'
      return line.split(delim)[1]
  raise RuntimeError('Unable to find version string.')


with open('README.md') as f:
  README = f.read()


setup(
  name='specialsauce',
  version=get_version(f'specialsauce/__init__.py'),
  description='Physiological formulas in Python',
  long_description=README,
  long_description_content_type='text/markdown',
  author='Aaron Schroeder',
  author_email='aaron@trailzealot.com',
  install_requires=[line.rstrip('\n') for line in open('requirements.txt')],
  url='https://github.com/aaron-schroeder/specialsauce',
  # project_urls={
  #   'Documentation': 'https://specialsauce.readthedocs.io/en/stable/',
  # },
  license='GPLv3',
  packages=find_packages(),
  classifiers=[
    'License :: OSI Approved :: MIT License',
    'Intended Audience :: Developers',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
  ]
)

